\documentclass[]{article}
\usepackage{lmodern}
\usepackage{amssymb,amsmath}
\usepackage{ifxetex,ifluatex}
\usepackage{fixltx2e} % provides \textsubscript
\ifnum 0\ifxetex 1\fi\ifluatex 1\fi=0 % if pdftex
  \usepackage[T1]{fontenc}
  \usepackage[utf8]{inputenc}
\else % if luatex or xelatex
  \ifxetex
    \usepackage{mathspec}
    \usepackage{xltxtra,xunicode}
  \else
    \usepackage{fontspec}
  \fi
  \defaultfontfeatures{Mapping=tex-text,Scale=MatchLowercase}
  \newcommand{\euro}{€}
\fi
% use upquote if available, for straight quotes in verbatim environments
\IfFileExists{upquote.sty}{\usepackage{upquote}}{}
% use microtype if available
\IfFileExists{microtype.sty}{%
\usepackage{microtype}
\UseMicrotypeSet[protrusion]{basicmath} % disable protrusion for tt fonts
}{}
\ifxetex
  \usepackage[setpagesize=false, % page size defined by xetex
              unicode=false, % unicode breaks when used with xetex
              xetex]{hyperref}
\else
  \usepackage[unicode=true]{hyperref}
\fi
\hypersetup{breaklinks=true,
            bookmarks=true,
            pdfauthor={Håkon Robbestad Gylterud},
            pdftitle={Binary disjunction},
            colorlinks=true,
            citecolor=blue,
            urlcolor=blue,
            linkcolor=magenta,
            pdfborder={0 0 0}}
\urlstyle{same}  % don't use monospace font for urls
\setlength{\parindent}{0pt}
\setlength{\parskip}{6pt plus 2pt minus 1pt}
\setlength{\emergencystretch}{3em}  % prevent overfull lines
\setcounter{secnumdepth}{0}

\title{Binary disjunction}
\author{Håkon Robbestad Gylterud}
\date{}

\begin{document}
\maketitle

This module implements both disjoint union and truncated binary
disjunction. We use the HoTT library truncations.

\begin{code}
module Propositions.Disjunction where
open import lib.Base
open import lib.NType

open import lib.types.Truncation
open import lib.types.Sigma
open import lib.types.Pi
open import lib.types.TLevel
\end{code}

\subsubsection{Binary disjoint union}\label{binary-disjoint-union}

\begin{code}
module _ {i j} where

  data _+_ (A : Type i) (B : Type j) : Type (lmax i j) where
     left : A → A + B
     right : B → A + B

  cases : ∀ {k} {A : Type i} {B : Type j}
        →  {C : A + B → Type k}
        → ((a : A) → C (left a))
        → ((b : B) → C (right b))
        → (x : A + B) → C x
  cases f g (left a) = f a
  cases f g (right b) = g b
\end{code}

\subsubsection{Binary truncated
disjunction}\label{binary-truncated-disjunction}

\begin{code}
  _∨_ : (A : Type i) (B : Type j) → Type (lmax i j)
  A ∨ B = Trunc (0 -1) (A + B)

  ∨-elim : ∀ {k} {A : Type i} {B : Type j} {P : A ∨ B → Type k}
         → (p : (x : A ∨ B) → is-prop (P x)) (d : (a : A + B) → P [ a ])
         → Π (A ∨ B) P
  ∨-elim p d = Trunc-elim p d

  ∨-elim' : ∀ {k} {A : Type i} {B : Type j} {P : A ∨ B → Type k}
         → (p : (x : A ∨ B) → is-prop (P x)) 
         → (d : (a : A) → P [ left a ])
         → (d' : (b : B) → P [ right b ])
         → Π (A ∨ B) P
  ∨-elim' p d d' = Trunc-elim p (cases d d')

  ∨-is-prop : (A : Type i) (B : Type j) → is-prop (A ∨ B)
  ∨-is-prop A B = Trunc-level
\end{code}

\end{document}
