\documentclass[]{article}
\usepackage{lmodern}
\usepackage{amssymb,amsmath}
\usepackage{ifxetex,ifluatex}
\usepackage{fixltx2e} % provides \textsubscript
\ifnum 0\ifxetex 1\fi\ifluatex 1\fi=0 % if pdftex
  \usepackage[T1]{fontenc}
  \usepackage[utf8]{inputenc}
\else % if luatex or xelatex
  \ifxetex
    \usepackage{mathspec}
    \usepackage{xltxtra,xunicode}
  \else
    \usepackage{fontspec}
  \fi
  \defaultfontfeatures{Mapping=tex-text,Scale=MatchLowercase}
  \newcommand{\euro}{€}
\fi
% use upquote if available, for straight quotes in verbatim environments
\IfFileExists{upquote.sty}{\usepackage{upquote}}{}
% use microtype if available
\IfFileExists{microtype.sty}{%
\usepackage{microtype}
\UseMicrotypeSet[protrusion]{basicmath} % disable protrusion for tt fonts
}{}
\ifxetex
  \usepackage[setpagesize=false, % page size defined by xetex
              unicode=false, % unicode breaks when used with xetex
              xetex]{hyperref}
\else
  \usepackage[unicode=true]{hyperref}
\fi
\hypersetup{breaklinks=true,
            bookmarks=true,
            pdfauthor={Håkon Robbestad Gylterud},
            pdftitle={Truncated Existential Quantifier},
            colorlinks=true,
            citecolor=blue,
            urlcolor=blue,
            linkcolor=magenta,
            pdfborder={0 0 0}}
\urlstyle{same}  % don't use monospace font for urls
\setlength{\parindent}{0pt}
\setlength{\parskip}{6pt plus 2pt minus 1pt}
\setlength{\emergencystretch}{3em}  % prevent overfull lines
\setcounter{secnumdepth}{0}

\title{Truncated Existential Quantifier}
\author{Håkon Robbestad Gylterud}
\date{}

\begin{document}
\maketitle

This module implements the truncated existensial quantifier, using the
HoTT library truncations.

\begin{code}
module Propositions.Existensial where
open import lib.Base
open import lib.NType

open import lib.types.Truncation
open import lib.types.Sigma
open import lib.types.Pi
open import lib.types.TLevel
\end{code}

\subsubsection{Truncated Existensial
quantification}\label{truncated-existensial-quantification}

\begin{code}
module _ {i j} (A : Type i) (B : A → Type j) where

  ∃ : Type (lmax i j)
  ∃ = Trunc (0 -1) (Σ A B)

  ∃-elim : ∀ {j} {P : ∃ → Type j}
         → (p : (x : ∃) → is-prop (P x)) (d : (a : Σ A B) → P [ a ])
         → Π ∃ P
  ∃-elim p = Trunc-elim p

  ∃-is-prop : is-prop ∃
  ∃-is-prop = Trunc-level
\end{code}

\end{document}
