\documentclass[]{article}
\usepackage{lmodern}
\usepackage{amssymb,amsmath}
\usepackage{ifxetex,ifluatex}
\usepackage{fixltx2e} % provides \textsubscript
\ifnum 0\ifxetex 1\fi\ifluatex 1\fi=0 % if pdftex
  \usepackage[T1]{fontenc}
  \usepackage[utf8]{inputenc}
\else % if luatex or xelatex
  \ifxetex
    \usepackage{mathspec}
    \usepackage{xltxtra,xunicode}
  \else
    \usepackage{fontspec}
  \fi
  \defaultfontfeatures{Mapping=tex-text,Scale=MatchLowercase}
  \newcommand{\euro}{€}
\fi
% use upquote if available, for straight quotes in verbatim environments
\IfFileExists{upquote.sty}{\usepackage{upquote}}{}
% use microtype if available
\IfFileExists{microtype.sty}{%
\usepackage{microtype}
\UseMicrotypeSet[protrusion]{basicmath} % disable protrusion for tt fonts
}{}
\ifxetex
  \usepackage[setpagesize=false, % page size defined by xetex
              unicode=false, % unicode breaks when used with xetex
              xetex]{hyperref}
\else
  \usepackage[unicode=true]{hyperref}
\fi
\hypersetup{breaklinks=true,
            bookmarks=true,
            pdfauthor={Håkon Robbestad Gylterud},
            pdftitle={The Identity Type of W-types -- result by Niels Anders Danielsson},
            colorlinks=true,
            citecolor=blue,
            urlcolor=blue,
            linkcolor=magenta,
            pdfborder={0 0 0}}
\urlstyle{same}  % don't use monospace font for urls
\setlength{\parindent}{0pt}
\setlength{\parskip}{6pt plus 2pt minus 1pt}
\setlength{\emergencystretch}{3em}  % prevent overfull lines
\setcounter{secnumdepth}{0}

\title{The Identity Type of W-types -- result by Niels Anders Danielsson}
\author{Håkon Robbestad Gylterud}
\date{}

\begin{document}
\maketitle

\begin{code}
{-# OPTIONS --without-K #-}
module W.W where

open import lib.Base
open import lib.Equivalences
\end{code}

\subsubsection{W-types}\label{w-types}

This module defines W-types and characterizes their identity type.

\begin{code}
data W {i j} {A : Type i} (B : A → Type j) : Type (lmax i j) where
 sup : (a : A) → (B a → W B) → W B

shape : ∀ {i j} {A : Type i} {B : A → Type j} → W B → A
shape (sup a _) = a

subtree : ∀ {i j} {A : Type i} {B : A → Type j} → (x : W B) → B (shape x) → W B
subtree (sup _ f) = f

open W
\end{code}

This is the unpacked equality on W-types:

\begin{code}
W-unpacked-id : forall {i j} {A : Type i} {B : A → Type j} 
              → (x y : W B) → Type (lmax i j)
W-unpacked-id {B = B} (sup a f) (sup a' f') 
    = Σ (a == a') 
        (\α → (f' ∘ (transport B α)) == f )
\end{code}

Home helper functions

\begin{code}
id-to-W-helper : forall {i j} {A : Set i} (B : A → Set j)
               → {a a' : A} → {f : B a → W B} {f' : B a' → W B}
               → (sup a f) == (sup a' f') → W-unpacked-id (sup a f) (sup a' f')
id-to-W-helper B idp = (idp , idp)

id-to-W : forall {i j} {A : Set i} (B : A → Set j) → {x y : W B}
              → x == y → W-unpacked-id x y
id-to-W B {x = (sup a f)} {y = (sup a' f')} p = id-to-W-helper B p

W-to-id-helper : forall {i j} {A : Set i} (B : A → Set j)
               → {a a' : A} → {f : B a → W B} {f' : B a' → W B}
               → W-unpacked-id (sup a f) (sup a' f') 
               → (sup a f) == (sup a' f') 
W-to-id-helper B (idp , idp) = idp


W-to-id : forall {i j} {A : Set i} (B : A → Set j) 
        → {x y : W B}
        → W-unpacked-id x y → x == y
W-to-id B {x = sup a f} {y = sup a' f'} α = W-to-id-helper B α
\end{code}

The main equivalence:

\begin{code}
W-id-≃ : forall {i j} {A : Set i} (B : A → Set j) → {x y : W B}
              → (x == y) ≃ W-unpacked-id x y
W-id-≃ {A = A} B {sup a f} {sup a' f'} = 
  ((id-to-W B) , is-eq _ (W-to-id B) (ε {f = f} {f' = f'}) δ) where
   ε : {a a' : A} {f : B a → W B} {f' : B a' → W B} 
      → (b : W-unpacked-id (sup a f) (sup a' f')) 
      → id-to-W B (W-to-id B {sup a f} {sup a' f'} b) == b
   ε (idp , idp) = idp

   δ : {a a' : A} {f : B a → W B} {f' : B a' → W B} 
      → (a : (sup a f) == (sup a' f')) 
      → W-to-id B (id-to-W B a) == a
   δ idp = idp
\end{code}

\subsubsection{W-induction}\label{w-induction}

\begin{code}
W-induction : ∀ {i j k} {A : Set i} (B : A → Set j)
            → (P : W B → Set k)
            → ((x : W B) → ((i : B (shape x)) → P (subtree x i)) → P x)
            → (x : W B) → P x
W-induction B P φ (sup a f) = φ (sup a f) (\i → W-induction B P φ (f i))
\end{code}

\end{document}
